import flet as ft
from fluxo.settings import AppSettings, AppThemeColors


class Footer(ft.UserControl):
    def __init__(self):
        super().__init__()

    def build(self):
        return ft.Container(
            content=ft.Row(
                controls=[
                    ft.Text(
                        value=AppSettings.VERSION,
                        color=AppThemeColors.BLACK_SECONDARY
                    )
                ],
                alignment=ft.MainAxisAlignment.CENTER
            ),
            padding=ft.padding.all(10)
        )


import flet as ft
import asyncio
from datetime import timedelta
from fluxo.settings import AppThemeColors
from fluxo.uttils import convert_str_to_datetime
from fluxo.fluxo_server.screens.app_bar import AppBar
from fluxo.fluxo_server.screens.footer import Footer
from fluxo.fluxo_core.database.task import ModelTask
from fluxo.fluxo_core.database.flow import ModelFlow
from fluxo.fluxo_core.database.log_execution_flow import ModelLogExecutionFlow


class Task(ft.UserControl):
    def __init__(self, task_id: int):
        super().__init__()
        self.task_id = task_id

    def build(self):
        self.text_name_task = ft.Ref[ft.Text]()
        self.text_name_flow = ft.Ref[ft.Text]()
        self.container_execution = ft.Ref[ft.Container]()

        self.text_start_time = ft.Ref[ft.Text]()
        self.text_end_time = ft.Ref[ft.Text]()
        self.text_duration = ft.Ref[ft.Text]()
        self.text_error = ft.Ref[ft.Text]()

        return ft.Column(
            controls=[
                ft.Container(
                    content=ft.ResponsiveRow(
                        controls=[
                            ft.Row(
                                controls=[
                                    ft.IconButton(
                                        icon=ft.icons.ARROW_BACK_ROUNDED,
                                        on_click=self.iconbutton_go_back,
                                        icon_color=AppThemeColors.BLACK,
                                        icon_size=30
                                    ),
                                    ft.Text(
                                        ref=self.text_name_flow,
                                        weight=ft.FontWeight.BOLD,
                                        color=AppThemeColors.BLACK,
                                        size=20
                                    ), # Text
                                    ft.Text(' > ', size=20, color=AppThemeColors.BLACK_SECONDARY), # Text
                                    ft.Text(
                                        ref=self.text_name_task,
                                        weight=ft.FontWeight.BOLD,
                                        color=AppThemeColors.BLACK,
                                        size=15
                                    ), # Text
                                    ft.Container(
                                        ref=self.container_execution,
                                        height=15,
                                        width=15,
                                        border_radius=ft.border_radius.all(15),
                                    ), # Container
                                ], # controls
                            ), # Row
                            ft.Container(height=10), # Container
                            ft.Container(
                                content=ft.Row(
                                    controls=[
                                        ft.Text(
                                            ref=self.text_start_time,
                                            color=AppThemeColors.BLACK,
                                            size=15
                                        ), # Text
                                    ], # controls
                                ), # Row
                                padding=ft.padding.only(left=50)
                            ),
                            ft.Container(
                                content=ft.Row(
                                    controls=[
                                        ft.Text(
                                            ref=self.text_end_time,
                                            color=AppThemeColors.BLACK,
                                            size=15
                                        ), # Text
                                    ], # controls
                                ), # Row
                                padding=ft.padding.only(left=50)
                            ),
                            ft.Container(
                                content=ft.Row(
                                    controls=[
                                        ft.Text(
                                            ref=self.text_duration,
                                            color=AppThemeColors.BLACK,
                                            size=15
                                        ), # Text
                                    ], # controls
                                ), # Row
                                padding=ft.padding.only(left=50)
                            ),
                            ft.Container(
                                content=ft.Row(
                                    controls=[
                                        ft.Text(
                                            value='Error:',
                                            color=AppThemeColors.BLACK,
                                            size=15
                                        ), # Text
                                        ft.Container(
                                            content=ft.Column(
                                                controls=[
                                                    ft.Text(
                                                        ref=self.text_error,
                                                        color=AppThemeColors.BLACK
                                                    ), # Text
                                                ], # controls
                                                scroll=ft.ScrollMode.ALWAYS,
                                            ), # Column
                                            width=850,
                                            height=300,
                                            bgcolor=AppThemeColors.GREY,
                                            border_radius=ft.border_radius.all(15),
                                            padding=ft.padding.all(10)
                                        ), # Container
                                    ], # controls
                                ), # Row
                                padding=ft.padding.only(left=50)
                            ),
                        ], # controls
                    ), # ResponsiveRow
                    width=900
                ), # Container
                ft.Container(padding=ft.padding.all(0)), # Container
            ], # controls
            scroll=ft.ScrollMode.AUTO,
            horizontal_alignment=ft.CrossAxisAlignment.CENTER
        ) # Column
    
    async def _load_attributes_task(self):
        task = ModelTask.get_by_id(self.task_id)
        flow = ModelFlow.get_by_id(task.flow_id)

        self.text_name_flow.current.value = flow.name
        self.text_name_task.current.value = task.name

        if task.end_time is None:
            self.container_execution.current.bgcolor = AppThemeColors.BLUE
            self.text_start_time.current.value = f'Start time: {task.start_time}'
            self.text_end_time.current.value = f'End time: -'
            self.text_duration.current.value = f'Duration: -'

        elif task.end_time and task.error:
            data_start_time = convert_str_to_datetime(task.start_time)
            data_execution_date = convert_str_to_datetime(task.execution_date)
            diference = data_execution_date - data_start_time

            self.container_execution.current.bgcolor = AppThemeColors.RED
            self.text_start_time.current.value = f'Start time: {task.start_time}'
            self.text_end_time.current.value = f'End time: {task.end_time}'
            self.text_duration.current.value = f'Duration: {diference.total_seconds()} seconds'
            self.text_error.current.value = task.error

        else:
            data_start_time = convert_str_to_datetime(task.start_time)
            data_execution_date = convert_str_to_datetime(task.execution_date)
            diference = data_execution_date - data_start_time

            self.container_execution.current.bgcolor = AppThemeColors.GREEN
            self.text_start_time.current.value = f'Start time: {task.start_time}'
            self.text_end_time.current.value = f'End time: {task.end_time}'
            self.text_duration.current.value = f'Duration: {diference.total_seconds()} seconds'
            self.text_error.current.value = task.error

        await self.update_async()

    async def iconbutton_go_back(self, e):
        log_flow_id = self.page.session.get('log_flow_id')
        await self.page.go_async(f'flow-execution/{log_flow_id}')

    async def did_mount_async(self):
        self.task_load_attributes_task = asyncio.create_task(self._load_attributes_task())

    async def will_unmount_async(self):
        self.task_load_attributes_task.cancel()
    

def view_task(task_id: int):

    return ft.View(
        route='task/:id',
        controls=[
            AppBar(),
            ft.Container(
                content=Task(task_id),
                expand=True,
                padding=ft.padding.all(15)
            ),
            Footer()
        ],
        padding=ft.padding.all(0),
        bgcolor=AppThemeColors.WHITE,
    )


import flet as ft
import asyncio
from fluxo.settings import AppThemeColors
from fluxo.fluxo_core.flows_executor import FlowsExecutor
from fluxo.fluxo_core.database.flow import ModelFlow
from fluxo.fluxo_core.database.task import ModelTask
from fluxo.fluxo_core.database.log_execution_flow import ModelLogExecutionFlow
from fluxo.fluxo_server.screens.home.status_execution import StatusExecution


class Flow(ft.UserControl):
    def __init__(self, flow: ModelFlow, flows_executor: FlowsExecutor):
        super().__init__()
        self.flow = flow
        self.flows_executor = flows_executor

    def build(self):
        self.row_flow = ft.Ref[ft.Row]()
        self.text_name = ft.Ref[ft.Text]()
        self.text_interval = ft.Ref[ft.Text]()

        self.row_executions = ft.Ref[ft.Row]()
        self.switch_running = ft.Ref[ft.Switch]()

        self.iconbutton_delete = ft.Ref[ft.IconButton]()
        self.iconbutton_run_now = ft.Ref[ft.IconButton]()

        return ft.Container(
            content=ft.Row(
                ref=self.row_flow,
                controls=[
                    ft.Container(
                        content=ft.Stack(
                            controls=[
                                ft.Container(
                                    content=ft.Text(
                                        ref=self.text_name,
                                        weight=ft.FontWeight.BOLD,
                                        color=AppThemeColors.BLACK,
                                        size=13,
                                        width=200
                                    ), # Text
                                    left=0,
                                    top=12
                                ), # Container
                                ft.Container(
                                    content=ft.Text(
                                        ref=self.text_interval,
                                        weight=ft.FontWeight.BOLD,
                                        color=AppThemeColors.BLACK_SECONDARY,
                                        size=10,
                                        width=200
                                    ), # Text
                                    left=150,
                                    top=0,
                                ), # Container
                            ], # controls
                        ), # Stack
                        width=290,
                        height=45,
                    ), # Container
                    ft.Container(
                        content=ft.Row(
                            ref=self.row_executions,
                            controls=[

                            ], # controls
                            width=320,
                            height=50,
                            scroll=ft.ScrollMode.AUTO,
                            auto_scroll=True
                        ), # Row
                    ), # Container
                    ft.Container(width=50),
                    ft.Tooltip(
                        message='Schedule Execution',
                        content=ft.Switch(
                            ref=self.switch_running,
                            label_position=ft.LabelPosition.LEFT,
                            on_change=self.on_change_switch_running
                        ),
                        bgcolor=AppThemeColors.QUARTENARY
                    ),
                    ft.Tooltip(
                        message='Run Now',
                        content=ft.IconButton(
                            ref=self.iconbutton_run_now,
                            icon=ft.icons.PLAY_ARROW_ROUNDED,
                            icon_color=AppThemeColors.BLACK_TERTIARY,
                            on_click=self.on_click_iconbutton_run_now
                        ),
                        bgcolor=AppThemeColors.QUARTENARY
                    ),
                    ft.Tooltip(
                        message='Delete Flow',
                        content=ft.IconButton(
                            ref=self.iconbutton_delete,
                            icon=ft.icons.DELETE_ROUNDED,
                            icon_color=AppThemeColors.BLACK_TERTIARY,
                            on_click=self.on_click_iconbutton_delete_flow
                        ),
                        bgcolor=AppThemeColors.QUARTENARY
                    )
                ], # controls
                #scroll=ft.ScrollMode.AUTO
            ), # Row
            bgcolor=AppThemeColors.GREY,
            border_radius=ft.border_radius.all(10),
            width=900,
            height=60,
            padding=ft.padding.only(left=15, top=0, right=15, bottom=0)
        ) # Container
    
    async def _load_attributes_flow(self):
        # Name
        self.text_name.current.value = self.flow.name

        # Interval
        interval = self.flow.interval
        if interval.get('minutes'):
            self.text_interval.current.value = f'every {interval.get("minutes")} min(s) at {interval.get("at")[1:]}s'
        elif interval.get('hours'):
            self.text_interval.current.value = f'every {interval.get("hours")} hour(s) at {interval.get("at")[1:]}m'
        elif interval.get('days'):
            self.text_interval.current.value = f'every {interval.get("days")} day(s) at {interval.get("at")}'
        
        # switch_start_stop_flow
        if self.flow.running:
            self.switch_running.current.value = True
            self.switch_running.current.label = 'ON'

            self.iconbutton_run_now.current.disabled = True
            self.iconbutton_run_now.current.icon = ft.icons.STOP_ROUNDED
            self.iconbutton_delete.current.disabled = True
        else:
            self.switch_running.current.value = False
            self.switch_running.current.label = 'OFF'

            self.iconbutton_run_now.current.disabled = False
            self.iconbutton_run_now.current.icon = ft.icons.PLAY_ARROW_ROUNDED
            self.iconbutton_delete.current.disabled = False
        await self.update_async()

    async def _load_status_executions(self):
        log_flows = ModelLogExecutionFlow.get_all_by_id_flow(self.flow.id)

        if log_flows:
            # Organize task list by date
            sorted_log_flows = sorted(log_flows, key=lambda x: x.date_of_creation, reverse=False)

            for log_flow in sorted_log_flows:
                self.row_executions.current.controls.append(StatusExecution(log_flow))
                        
            if len(log_flows) < 10:
                n = 10 - len(log_flows)
                for _ in range(n):
                    self.row_executions.current.controls.insert(
                        0,
                        ft.Container(
                            bgcolor=AppThemeColors.WHITE,
                            height=23,
                            width=23,
                            border_radius=ft.border_radius.all(15),
                            tooltip=''
                        ), # Container
                    )
        else:
            for _ in range(10):
                self.row_executions.current.controls.insert(
                    0,
                    ft.Container(
                        bgcolor=AppThemeColors.WHITE,
                        height=23,
                        width=23,
                        border_radius=ft.border_radius.all(15),
                        tooltip=''
                    ), # Container
                )
        await self.update_async()

    async def on_change_switch_running(self, e):
        flow = ModelFlow.get_by_id(self.flow.id)

        if flow.running: # Stop Flow
            self.flows_executor.stop_flow_execution([flow])
            e.control.value = False
            e.control.label = 'OFF'

            self.iconbutton_run_now.current.disabled = False
            self.iconbutton_run_now.current.icon = ft.icons.PLAY_ARROW_ROUNDED
            self.iconbutton_delete.current.disabled = False

        else: # Start Flow
            self.flows_executor.execute_parallel_flows([flow])
            e.control.value = True
            e.control.label = 'ON'

            self.iconbutton_run_now.current.disabled = True
            self.iconbutton_run_now.current.icon = ft.icons.STOP_ROUNDED
            self.iconbutton_delete.current.disabled = True
        await self.update_async()

    async def on_click_iconbutton_run_now(self, e):
        self.flows_executor.execute_flow_now([self.flow])

        e.control.disabled = True
        e.control.icon = ft.icons.STOP_ROUNDED
        self.iconbutton_delete.current.disabled = True  
        
        await self.update_async()

    async def on_click_iconbutton_delete_flow(self, e):
        e.control.disabled = True
        await self.update_async()

        ModelFlow.delete(self.flow.id)

        list_log_flow = ModelLogExecutionFlow.get_all_by_id_flow(self.flow.id)
        if list_log_flow:
            for log_flow in list_log_flow:
                ModelLogExecutionFlow.delete(log_flow.id)

        list_task = ModelTask.get_all_by_fluxo_id(self.flow.id)
        if list_task:
            for task in list_task:
                ModelTask.delete(task.id)
        await self.clean_async()

    async def did_mount_async(self):
        self.task_load_attributes_flow = asyncio.create_task(self._load_attributes_flow())
        self.task_load_status_executions = asyncio.create_task(self._load_status_executions())

    async def will_unmount_async(self):
        self.task_load_attributes_flow.cancel()
        self.task_load_status_executions.cancel()


import os
import atexit
import multiprocessing
import importlib
import asyncio
import schedule
import signal
from time import sleep
from typing import List, Optional
from fluxo.settings import PathFilesPython, Db
from fluxo.uttils import current_time_formatted
from fluxo.logging import logger
from fluxo.fluxo_core.database.db import _verify_if_db_exists
from fluxo.fluxo_core.database.app import ModelApp
from fluxo.fluxo_core.database.flow import ModelFlow
from fluxo.fluxo_core.database.task import ModelTask
from fluxo.fluxo_core.database.log_execution_flow import ModelLogExecutionFlow


class FlowsExecutor:
    '''
    Represents a 'FlowsExecutor' object responsible for executing Flow files in parallel processes.

    Attributes:
        - path (str): The path to the directory containing Fluxo files.

    Methods:
        - execute_parallel_flows(): Executes Flow files in parallel processes.
        - stop_flow_execution(): Stops the execution of the specified flows.
    '''
    _coroutines: list = [] # [(task, flow_info)]

    def __init__(self, path=PathFilesPython.PATH_FILES_PYTHON) -> None:
        '''
        Initializes a new 'FlowsExecutor' instance.

        The instance is configured with the path to the directory containing Flow files,
        an empty list to store multiprocessing.Process instances, and the verification
        of the existence of the database.
        '''
        self.path = path
        self.processes = []

        # Register the functions to be executed on program exit
        atexit.register(self._cleanup_processes)
        atexit.register(FlowsExecutor._change_app_status_to_false)
  
    def execute_parallel_flows(self, flows: Optional[List[ModelFlow]] = None):
        '''
        Executes Flow files in parallel processes.
        '''
        # Check if the database doesn't exist
        if not self._db_exists():
            # Verify and create the database if it doesn't exist
            _verify_if_db_exists()
            FlowsExecutor._change_app_status_to_true() # Change status to True in database

            for file in os.listdir(self.path):
                if file.endswith(".py"):
                    process = multiprocessing.Process(
                        target=FlowsExecutor._execute_async_tasks_first_time, args=(self.path, file))
                    self.processes.append(process)
                    process.start()
        else:
            FlowsExecutor._change_app_status_to_true() # Change status to True in database
            # If flows is None, then all flows will be executed
            if flows is None:
                all_flows = ModelFlow.get_all()
                for flow in all_flows:
                    process = multiprocessing.Process(
                        target=FlowsExecutor._execute_async_tasks_from_flow, args=(self.path, flow))
                    self.processes.append(process)
                    process.start()
            else:
                for flow in flows:
                    process = multiprocessing.Process(
                        target=FlowsExecutor._execute_async_tasks_from_flow, args=(self.path, flow))
                    self.processes.append(process)
                    process.start()

    def execute_flow_now(self, flows: Optional[List[ModelFlow]] = None):
        '''
        Executes Flow right now.
        '''
        for flow in flows:
            process = multiprocessing.Process(
                target=FlowsExecutor._execute_async_tasks_now, args=(self.path, flow))
            self.processes.append(process)
            process.start()
                    
    def update_new_flow_in_python_files(self):
        # Check if the database doesn't exist
        if not self._db_exists():
            # Verify and create the database if it doesn't exist
            _verify_if_db_exists()
            
        for file in os.listdir(self.path):
            if file.endswith(".py"):
                FlowsExecutor._update_tasks_in_new_flow(self.path, file)

    def stop_flow_execution(self, flows: List[ModelFlow]):
        '''
        Stops the execution of the specified flows.

        Parameters:
            - flows (List[ModelFlow]): The list of ModelFlow objects representing flows to be stopped.

        This method iterates through the provided list of flows and attempts to stop their execution.
        If a flow is currently running, it terminates the associated process using the process ID (PID).
        If the process is not found or if there are permission issues, exceptions are raised.
        
        Note: This method updates the running status and process information in the database.
        '''
        for flow in flows:
            flow = ModelFlow.get_by_id(flow.id)
            if flow:
                if flow.running:
                    pid = flow.running_process.get('process_pid')
                    try:
                        flow.running_process = None
                        flow.running = False
                        flow.update(**flow.__dict__)
                        logger.info(f'Flow [{flow.name}] execution scheduling canceled')
                        os.kill(pid, signal.SIGTERM)
                    except ProcessLookupError:
                        raise Exception(f'The process with PID {pid} was not found')
                    except PermissionError:
                        raise Exception(f'You are not allowed to terminate the process with PID {pid}.')
  
    @staticmethod
    def _execute_async_tasks_from_flow(path, flow: Optional[ModelFlow] = None):
        '''
        Executes asynchronous tasks defined in a module.

        Parameters:
            - path (str): The path to the directory containing the module.
            - flow (str): Flow and its tasks to be executed.
        '''
        for file in os.listdir(path):
            if file.endswith('.py'):
                try:
                    # Remove the '.py' extension to get the module name
                    name_module = file[:-3]

                    # Dynamically import the module
                    dir_base = os.path.basename(path)
                    module = importlib.import_module(f"{dir_base}.{name_module}")

                    # Search for asynchronous functions in the module
                    for name_attribute in dir(module):
                        attribute = getattr(module, name_attribute)
                        if asyncio.iscoroutinefunction(attribute):
                            # Check if the function is decorated with @Task
                            if hasattr(attribute, 'task_info'):

                                task_info = attribute.task_info
                                flow_info = task_info.get('flow')
                                flow_register_db = ModelFlow.get_by_name(flow_info.name)

                                if flow_register_db is None:  # Check if fluxo exists in the database
                                    new_flow = ModelFlow(
                                        name=flow_info.name,
                                        interval=flow_info.interval,
                                        list_names_tasks=[task_info.get('name')],
                                        running=False
                                    )
                                    flow = new_flow.save()
                                    logger.info(f'New Flow [{new_flow.name}] update in database')

                                if flow.name == flow_info.name:
                                    flow.interval = flow_info.interval
                                    flow.active = flow_info.active
                                    flow.update(**flow.__dict__)
                                    
                                    if flow.active: # Check if fluxo.active is True
                                        FlowsExecutor._coroutines.append((attribute, flow_info))

                    if FlowsExecutor._coroutines:
                        flow.running = True
                        # Sets the running process to the flow, storing the PID of the current process.
                        flow.running_process = {'process_pid': os.getpid()}
                        flow.update(**flow.__dict__)
                        logger.info(f'Flow [{flow.name}] execution scheduling started')
                        FlowsExecutor._schedule_async_tasks()

                except Exception as e:
                    raise Exception(f"Error executing task in module")
                
    @staticmethod
    def _execute_async_tasks_now(path, flow: Optional[ModelFlow] = None):
        '''
        Executes asynchronous tasks defined in a module.

        Parameters:
            - path (str): The path to the directory containing the module.
            - flow (str): Flow and its tasks to be executed.
        '''
        for file in os.listdir(path):
            if file.endswith('.py'):
                try:
                    # Remove the '.py' extension to get the module name
                    name_module = file[:-3]

                    # Dynamically import the module
                    dir_base = os.path.basename(path)
                    module = importlib.import_module(f"{dir_base}.{name_module}")

                    # Search for asynchronous functions in the module
                    for name_attribute in dir(module):
                        attribute = getattr(module, name_attribute)
                        if asyncio.iscoroutinefunction(attribute):
                            # Check if the function is decorated with @Task
                            if hasattr(attribute, 'task_info'):

                                task_info = attribute.task_info
                                flow_info = task_info.get('flow')
                                flow_register_db = ModelFlow.get_by_name(flow_info.name)

                                if flow_register_db is None:  # Check if fluxo exists in the database
                                    new_flow = ModelFlow(
                                        name=flow_info.name,
                                        interval=flow_info.interval,
                                        list_names_tasks=[task_info.get('name')],
                                        running=False
                                    )
                                    flow = new_flow.save()
                                    logger.info(f'New Flow [{new_flow.name}] update in database')

                                if flow.name == flow_info.name:
                                    flow.interval = flow_info.interval
                                    flow.active = flow_info.active
                                    flow.update(**flow.__dict__)
                                    
                                    if flow.active: # Check if fluxo.active is True
                                        FlowsExecutor._coroutines.append((attribute, flow_info))

                    if FlowsExecutor._coroutines:
                        flow.running = True
                        # Sets the running process to the flow, storing the PID of the current process.
                        flow.running_process = {'process_pid': os.getpid()}
                        flow.update(**flow.__dict__)
                        logger.info(f'Flow [{flow.name}] execution scheduling started')
                        FlowsExecutor._schedule_async_tasks_now()

                except Exception as e:
                    raise Exception(f"Error executing task in module")

    @staticmethod
    def _execute_async_tasks_first_time(path, file):
        '''
        Executes asynchronous tasks defined in a module for the first time.

        Parameters:
        - path (str): The path to the directory containing the module.
        - file (str): The name of the module file (excluding the '.py' extension).
        '''
        try:
            # Remove the '.py' extension to get the module name
            name_module = file[:-3]

            # Dynamically import the module
            dir_base = os.path.basename(path)
            module = importlib.import_module(f"{dir_base}.{name_module}")

            # Search for asynchronous functions in the module
            for name_attribute in dir(module):
                attribute = getattr(module, name_attribute)
                if asyncio.iscoroutinefunction(attribute):
                    # Check if the function is decorated with @Task
                    if hasattr(attribute, 'task_info'):

                        task_info = attribute.task_info
                        flow_info = task_info.get('flow')
                        flow = ModelFlow.get_by_name(flow_info.name)

                        if flow is None:  # Check if fluxo exists in the database
                            flow = ModelFlow(
                                name=flow_info.name,
                                interval=flow_info.interval,
                                list_names_tasks=[task_info.get('name')],
                                running=False
                            )
                            flow = flow.save()
                            logger.info(f'New Flow [{flow.name}] update in database')
                        else:
                            flow.name = flow_info.name # Update Flow in database
                            flow.interval = flow_info.interval
                            flow.list_names_tasks.append(task_info.get('name'))
                            flow.update(**flow.__dict__)

        except Exception as e:
            raise Exception(f"Error executing task in module")
        
    @staticmethod
    def _update_tasks_in_new_flow(path, file):
        '''
        Update asynchronous tasks defined in a new module for a new flow.

        Parameters:
        - path (str): The path to the directory containing the module.
        - file (str): The name of the module file (excluding the '.py' extension).
        '''
        try:
            # Remove the '.py' extension to get the module name
            name_module = file[:-3]

            # Dynamically import the module
            dir_base = os.path.basename(path)
            module = importlib.import_module(f"{dir_base}.{name_module}")

            _name_flow = None
            _list_names_tasks: list = []
            # Search for asynchronous functions in the module
            for name_attribute in dir(module):
                attribute = getattr(module, name_attribute)
                if asyncio.iscoroutinefunction(attribute):
                    # Check if the function is decorated with @Task
                    if hasattr(attribute, 'task_info'):

                        task_info = attribute.task_info
                        flow_info = task_info.get('flow')
                        flow = ModelFlow.get_by_name(flow_info.name)

                        if flow is None:  # Check if fluxo exists in the database
                            flow = ModelFlow(
                                name=flow_info.name,
                                interval=flow_info.interval,
                                list_names_tasks=[task_info.get('name')],
                                running=False
                            )
                            flow = flow.save()
                            _name_flow = flow
                            logger.info(f'New Flow [{flow.name}] update in database')
                        else:
                            if _name_flow:
                                _list_names_tasks.append(task_info.get('name'))
            if _name_flow:
                if _list_names_tasks:
                    for _tasks in _list_names_tasks:
                        _name_flow.list_names_tasks.append(_tasks)
                _name_flow.update(**_name_flow.__dict__)

        except Exception as e:
            raise Exception(f"Error update task in module.")

    @staticmethod
    def _schedule_async_tasks():
        '''
        Schedules the execution of asynchronous tasks based on the specified intervals.
        '''
        for task, flow_info in FlowsExecutor._coroutines:
            if flow_info.interval.get('minutes'):
                schedule.every(flow_info.interval.get('minutes')).minutes.at(
                    flow_info.interval.get('at')).do(FlowsExecutor._run_asynchronous_task, task)
            elif flow_info.interval.get('hours'):
                schedule.every(flow_info.interval.get('hours')).hours.at(
                    flow_info.interval.get('at')).do(FlowsExecutor._run_asynchronous_task, task)
            elif flow_info.interval.get('days'):
                schedule.every(flow_info.interval.get('days')).days.at(
                    flow_info.interval.get('at')).do(FlowsExecutor._run_asynchronous_task, task)
                
        jobs_pending = [schedule.get_jobs()]

        condition = True
        while condition:
            try:
                schedule.run_pending()
                sleep(1)
            except KeyboardInterrupt:
                for job in jobs_pending:
                    schedule.cancel_job(job)
                condition = False

                _tasks, flow = zip(*FlowsExecutor._coroutines)

                flow = flow[0]
                FlowsExecutor._update_tasks_in_db_if_keyboardinterrupt(flow.name)
                FlowsExecutor._stop_flow_execution(flow.name)

    @staticmethod
    def _schedule_async_tasks_now():
        '''
        Schedules the execution of asynchronous tasks right now.
        '''
        for task, flow_info in FlowsExecutor._coroutines:
            if flow_info.interval.get('minutes'):
                schedule.every(flow_info.interval.get('minutes')).minutes.at(
                    flow_info.interval.get('at')).do(FlowsExecutor._run_asynchronous_task, task)
            elif flow_info.interval.get('hours'):
                schedule.every(flow_info.interval.get('hours')).hours.at(
                    flow_info.interval.get('at')).do(FlowsExecutor._run_asynchronous_task, task)
            elif flow_info.interval.get('days'):
                schedule.every(flow_info.interval.get('days')).days.at(
                    flow_info.interval.get('at')).do(FlowsExecutor._run_asynchronous_task, task)
                
        jobs_pending = [schedule.get_jobs()]

        _tasks, flow = zip(*FlowsExecutor._coroutines)
        flow = flow[0]

        try:
            schedule.run_all()
            schedule.clear()
            FlowsExecutor._stop_flow_execution(flow.name)
        except KeyboardInterrupt:
            for job in jobs_pending:
                schedule.cancel_job(job)

            FlowsExecutor._update_tasks_in_db_if_keyboardinterrupt(flow.name)
            FlowsExecutor._stop_flow_execution(flow.name)

    def _cleanup_processes(self):
        '''
        Terminates all running processes.
        '''
        logger.info('Terminating all processes')
        try:
            for process in self.processes:
                process.join()
        except Exception as err:
            print(f'===> {err}')

    def _db_exists(self):
        '''
        Verifies if the database file exists at the specified path.

        Returns:
            True or None
        '''
        return os.path.exists(Db.PATH)
    
    @staticmethod
    def _update_tasks_in_db_if_keyboardinterrupt(flow_name):
        flow = ModelFlow.get_by_name(flow_name)
        tasks = ModelTask.get_all_by_fluxo_id(flow.id)

        for task in tasks:
            if task.end_time is None:
                task.error = 'KeyboardInterrupt'
                task.end_time = current_time_formatted()
                task.execution_date = task.end_time
                task.update(**task.__dict__)

    @staticmethod
    def _stop_flow_execution(flow_name: str):
        '''
        Stops the execution of the specified flow.

        Parameters:
            - flow_name: Name of ModelFlow objects representing flow to be stopped.

        This method iterates through the provided flow_name and attempts to stop their execution.
        If a flow is currently running, it terminates the associated process using the process ID (PID).
        If the process is not found or if there are permission issues, exceptions are raised.
        
        Note: This method updates the running status and running_process information in the database.
        '''
        flow = ModelFlow.get_by_name(flow_name)
        log_flow = ModelLogExecutionFlow.get_by_idflow_and_endtime_is_none(flow.id)
        if flow:
            if flow.running:
                pid = flow.running_process.get('process_pid')
                try:
                    flow.running_process = None
                    flow.running = False
                    flow.update(**flow.__dict__)
                    if log_flow:
                        log_flow.delete(log_flow.id)
                        
                    logger.info(f'Flow [{flow.name}] execution scheduling canceled')
                    os.kill(pid, signal.SIGTERM)
                except ProcessLookupError:
                    raise Exception(f'The process with PID {pid} was not found')
                except PermissionError:
                    raise Exception(f'You are not allowed to terminate the process with PID {pid}.')

    @staticmethod
    def _change_app_status_to_true():
        '''
        Changes the application status to True in the database.

        If the application already exists, its status is updated to True.
        Otherwise, a new application with active status set to True is created and saved.
        '''
        app = ModelApp.get()
        if app:
            app.update(app.id, True)
        else:
            app = ModelApp(active=True)
            app.save()

    @staticmethod
    def _change_app_status_to_false():
        '''
        Changes the application status to False in the database.

        If the application exists, its status is updated to False.
        '''
        app = ModelApp.get()
        if app:
            app.update(app.id, False)

    @staticmethod
    def _run_asynchronous_task(task):
        '''
        Runs an asynchronous task.

        Parameters:
        - task: The asynchronous task to be executed.
        '''
        asyncio.run(task())



from datetime import datetime, timedelta


def convert_str_to_datetime(data_str: str):
    if data_str is None:
        return None
    else:
        return datetime.strptime(data_str, "%Y/%m/%d %H:%M:%S")
    
def convert_datetime_to_str(dt_obj: datetime):
    if dt_obj is None:
        return None
    else:
        return dt_obj.strftime("%Y/%m/%d %H:%M:%S")

  
def current_time_formatted():
    '''
    Returns the current time formatted as a string.

    Returns:
    str: The formatted current time.
    '''
    current_time = datetime.now()
    format_str = "%Y/%m/%d %H:%M:%S"
    formatted_time = current_time.strftime(format_str)
    return formatted_time


def add_minutes_to_utc_time(minutes: int, time=None):
    '''
    Adds a specified number of minutes to a given time or the current time.

    Parameters:
    - minutes (int): The number of minutes to add.
    - time (str): The time to which minutes should be added. Defaults to None.

    Returns:
    str: The formatted time after adding the specified minutes.
    '''
    if time is None:
        time = current_time_formatted()

    # Convert a string to a datetime object
    time = datetime.strptime(time, "%Y/%m/%d %H:%M:%S")

    new_time = time + timedelta(minutes=minutes)
    format_str = "%Y/%m/%d %H:%M:%S"
    formatted_time = new_time.strftime(format_str)
    return formatted_time



import flet as ft
from fluxo.settings import FONTS, AppThemeColors, AppSettings
from fluxo.fluxo_server.screens.home.home import view_home
from fluxo.fluxo_server.screens.flow_execution.flow_execution import view_flow_execution
from fluxo.fluxo_server.screens.task.task import view_task


class App:
    def __init__(self, page: ft.Page) -> None:
        self.page = page

    async def init(self):
        self.page.theme = ft.Theme(
            color_scheme_seed=AppThemeColors.PRIMARY,
        )
        self.page.fonts = FONTS
        self.page.title = 'Fluxo'

        async def route_change(route):
            #self.page.views.clear()
            t_route = ft.TemplateRoute(self.page.route)

            if t_route.match('/'):
                self.page.views.append(view_home())
            
            elif t_route.match('flow-execution/:id'):
                self.page.views.append(view_flow_execution(t_route.id))

            elif t_route.match('task/:id'):
                self.page.views.append(view_task(t_route.id))

            await self.page.update_async()

        async def view_pop(view):
            self.page.views.pop()
            top_view = self.page.views[-1]
            await self.page.go_async(top_view.route)

        self.page.on_route_change = route_change
        self.page.on_view_pop = view_pop
        await self.page.go_async(self.page.route)


async def main(page: ft.Page):
    await App(page).init()

if __name__ == '__main__':
    ft.app(target=main, 
           view=ft.AppView.WEB_BROWSER,
           port=AppSettings.PORT,
           assets_dir=AppSettings.ASSETS_DIR)




class Flow:
    '''
    Represents a 'Flow' object with attributes such as name, interval, and active status.

    Attributes:
        - name (str): The name of the 'Flow'.
        - interval (dict): The interval information for the 'Flow'.
        - active (bool): A flag indicating whether the 'Flow' is active or not.

    Example:
        ```
        from fluxo import Flow, Task, Minutes

        interval = Minutes(1, 30).format()
        flow = Flow(name='My Flow 1', interval=interval)

        @Task('My Task 1', flow=flow)
        async def My_func():
            print('My_func executed!')
        ```
    '''
    def __init__(
        self,
        name: str,
        interval: dict = None,
        active: bool = True
    ):
        self.name = name
        self.interval = interval
        self.active = active



import flet as ft
import os


class AppSettings:
    '''Configurações da aplicação Fluxo'''
    VERSION = 'v0.14.2'
    PORT = 7777
    ASSETS_DIR = 'fluxo_server/assets'

class AppThemeColors:
    '''Cores do tema da aplicação Fluxo'''
    PRIMARY = '#6929F2'
    SECONDARY = '#8F49F2'
    TERTIARY = '#BC6BF2'
    QUARTENARY = '#BFA7F2'
    BLACK = ft.colors.BLACK
    BLACK_SECONDARY = ft.colors.BLACK87
    BLACK_TERTIARY = ft.colors.BLACK54
    GREY = '#F2F2F2'
    WHITE = ft.colors.WHITE
    WHITE_SECONDARY = ft.colors.WHITE70
    GREEN = ft.colors.GREEN
    RED = ft.colors.RED
    BLUE = ft.colors.BLUE

class Db:
    NAME = 'database_fluxo.sqlite3'
    PATH = os.path.join(os.getcwd(), NAME)

class PathFilesPython:
    FOLDER = 'python_files'
    PATH_FILES_PYTHON = os.path.join(os.getcwd(), FOLDER)

# Fontes
FONTS = {
    'Open Sans': '/fonts/OpenSans-Regular.ttf',
    'Open Sans Bold': '/fonts/OpenSans-Bold.ttf'
}



import sqlite3
import json
from datetime import datetime
from dataclasses import dataclass
from fluxo.settings import Db
from fluxo.uttils import current_time_formatted


@dataclass
class ModelFlow:
    '''
    Represents a 'Flow' object with attributes corresponding to the columns
    in the 'TB_Flow' table in the SQLite database.

    Attributes:
        - id (int): The unique identifier for the 'Fluxo'.
        - name (str): The name of the 'Fluxo'.
        - date_of_creation (datetime): The date and time when the 'Fluxo' was created.
        - interval (dict): The interval information for the 'Fluxo'. Ex `{'minutes':1, 'at':':10'}`
        - active (bool): A flag indicating whether the 'Fluxo' is active or not.
        - list_names_tasks: (list): List of task names linked to the flow.
        - running (bool): If the flow is running.
        - running_process (dict): What PID of the process the flow is running on.

    Methods:
        - save(): Saves the current 'Flow' instance to the 'TB_Flow' table in the database.
        - update(id, name, date_of_creation, interval, active): Updates the 'Flow' with the specified ID
                with the provided information in the 'TB_Flow' table.
        - get_all(): Retrieves all 'Flow' instances from the 'TB_Flow' table.
        - get_by_name(name): Retrieves a 'Flow' instance by its name from the 'TB_Flow' table.
        - get_by_id(id): Retrieves a 'Flow' instance by its ID from the 'TB_Flow' table.
        - delete(id): Deletes the 'Flow' with the specified ID from the 'TB_Flow' table.
    '''
    id: int = None
    name: str = None
    date_of_creation: datetime = None
    interval: dict = None
    active: bool = True
    list_names_tasks: list = None
    running: bool = None
    running_process: dict = None

    def save(self):
        '''
        Saves the current 'Fluxo' instance to the 'TB_Flow' table in the database.
        '''
        date_of_creation = current_time_formatted()
        interval = json.dumps(self.interval) if self.interval else None
        list_names_tasks = json.dumps(self.list_names_tasks) if self.list_names_tasks else None
        running_process = json.dumps(self.running_process) if self.running_process else None

        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('''
            INSERT INTO TB_Flow (name, date_of_creation, interval, active, list_names_tasks, running, running_process)
            VALUES (?, ?, ?, ?, ?, ?, ?)
        ''', (self.name, date_of_creation, interval, self.active, list_names_tasks, self.running, running_process))
        conn.commit()

        # Retrieve fluxo ID after insertion
        cursor.execute('SELECT last_insert_rowid()')
        fluxo_id = cursor.fetchone()[0]

        conn.close()

        return ModelFlow.get_by_id(fluxo_id)

    @staticmethod
    def update(id, name, date_of_creation, interval, active, list_names_tasks, running, running_process):
        '''
        Updates the 'Flow' with the specified ID with the provided information
        in the 'TB_Flow' table.

        Parameters:
        - id (int): The ID of the 'Fluxo' to be updated.
        - name (str): The new name for the 'Flow'.
        - date_of_creation (datetime): The new date and time of creation for the 'Fluxo'.
        - interval (dict): The new interval information for the 'Fluxo'.
        - active (bool): The new active status for the 'Fluxo'.
        - list_names_tasks: (list): The new List of task names linked to the flow.
        - running (bool): If the flow is running.
        - running_process (dict): What PID of the process the flow is running on.
        '''
        interval = json.dumps(interval) if interval else None
        list_names_tasks = json.dumps(list_names_tasks) if list_names_tasks else None
        running_process = json.dumps(running_process) if running_process else None

        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('''
            UPDATE TB_Flow
            SET name=?, date_of_creation=?, interval=?, active=?, list_names_tasks=?, running=?, running_process=?
            WHERE id=?
        ''', (name, date_of_creation, interval, active, list_names_tasks, running, running_process, id))
        conn.commit()
        conn.close()

    @staticmethod
    def get_all():
        '''
        Retrieves all 'Fluxo' instances from the 'TB_Flow' table.

        Returns:
        List[ModelFlow]: A list containing all 'Flow' instances in the database.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_Flow')
        data = cursor.fetchall()
        conn.close()

        flows = []
        for row in data:
            flow = ModelFlow(*row)
            
            # Converter strings JSON from interval, list_names_tasks and running_process to dicts
            flow.interval = json.loads(flow.interval) if flow.interval else None
            flow.list_names_tasks = json.loads(flow.list_names_tasks) if flow.list_names_tasks else None
            flow.running_process = json.loads(flow.running_process) if flow.running_process else None

            flows.append(flow)

        if flows:
            return flows
        else:
            return None

    @staticmethod
    def get_by_name(name):
        '''
        Retrieves a 'Flow' instance by its name from the 'TB_Flow' table.

        Parameters:
        - name (str): The name of the 'Flow' to be retrieved.

        Returns:
            Flow or None: The 'Flow' instance if found, or None if not found.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_Flow WHERE name=?', (name,))
        data = cursor.fetchone()
        conn.close()
        
        if data:
            flow = ModelFlow(*data)

            # Converter strings JSON from interval, list_names_tasks and running_process to dicts
            flow.interval = json.loads(flow.interval) if flow.interval else None
            flow.list_names_tasks = json.loads(flow.list_names_tasks) if flow.list_names_tasks else None
            flow.running_process = json.loads(flow.running_process) if flow.running_process else None

            return flow
        else:
            return None
        
    @staticmethod
    def get_by_id(id):
        '''
        Retrieves a 'Flow' instance by its ID from the 'TB_Flow' table.

        Parameters:
        - id (int): The ID of the 'Flow' to be retrieved.

        Returns:
            Flow or None: The 'Flow' instance if found, or None if not found.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_Flow WHERE id=?', (id,))
        data = cursor.fetchone()
        conn.close()

        if data:
            flow = ModelFlow(*data)

            # Converter strings JSON from interval, list_names_tasks and running_process to dicts
            flow.interval = json.loads(flow.interval) if flow.interval else None
            flow.list_names_tasks = json.loads(flow.list_names_tasks) if flow.list_names_tasks else None
            flow.running_process = json.loads(flow.running_process) if flow.running_process else None

            return flow
        else:
            return None

    @staticmethod
    def delete(id):
        '''
        Deletes the 'Flow' with the specified ID from the 'TB_Flow' table.

        Parameters:
            - id (int): The ID of the 'Flow' to be deleted.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('DELETE FROM TB_Flow WHERE id=?', (id,))
        conn.commit()
        conn.close()

    def __repr__(self) -> str:
        '''
        Returns a string representation of the 'Flow' instance.
        '''
        return f'''
            id:                     {self.id},
            name:                   {self.name},
            date_of_creation:       {self.date_of_creation},
            interval:               {self.interval},
            active:                 {self.active},
            list_names_tasks:       {self.list_names_tasks},
            running:                {self.running},
            running_process:        {self.running_process},
        '''



import os
import sqlite3
from fluxo.settings import Db
from fluxo.logging import logger


def _verify_if_db_exists(path_db: str = Db.PATH):
    '''
    Verifies if the database file exists at the specified path. If the file does not exist,
    it creates a new database using the `create_db` function.

    Parameters:
    - path_db (str): The path to the database file. Defaults to the path specified in the
      `Db.PATH` variable.

    Returns:
        None
    '''
    if not os.path.exists(path_db):
        create_db(path_db)


def create_db(path_db: str):
    '''
    Creates a SQLite database at the specified path. If the tables do not exist, 
    it creates them along with their respective columns.

    Parameters:
    - path_db (str): The path to the SQLite database file.

    Return:
    None
    '''
    try:
        # Database connection
        conn = sqlite3.connect(path_db)

        # Create TB_Flow table
        conn.execute('''
            CREATE TABLE IF NOT EXISTS TB_Flow (
                id INTEGER PRIMARY KEY,
                name TEXT,
                date_of_creation DATE,
                interval TEXT,
                active BOOLEAN,
                list_names_tasks TEXT, -- Storing the list as a JSON string
                running BOOLEAN,
                running_process TEXT -- Storing the list as a JSON string
            )
        ''')
        logger.info('TB_Flow table created successfully')

        # Create TB_Task table
        conn.execute('''
            CREATE TABLE IF NOT EXISTS TB_Task (
                id INTEGER PRIMARY KEY,
                name TEXT,
                execution_date DATE,
                flow_id INTEGER,
                start_time DATE,
                end_time DATE,
                error TEXT
            )
        ''')
        logger.info('TB_Task table created successfully')
        
        # Create TB_LogExecutionFlow table
        conn.execute('''
            CREATE TABLE TB_LogExecutionFlow (
                id INTEGER PRIMARY KEY,
                name TEXT,
                date_of_creation DATETIME,
                start_time DATETIME,
                end_time DATETIME,
                id_flow INTEGER,
                ids_task TEXT,  -- Storing the list as a JSON string
                ids_error_task TEXT  -- Storing the list as a JSON string
            )
        ''')
        logger.info('TB_LogExecutionFlow table created successfully')
        
        # Create TB_App table
        conn.execute('''
            CREATE TABLE IF NOT EXISTS TB_App (
                id INTEGER PRIMARY KEY,
                active BOOLEAN,
                active_since DATE
            )
        ''')
        logger.info('TB_App table created successfully')

    
    except Exception as err:
        # Log the error
        logger.error(f'Error during database creation: {err}')

        # Close the connection if open
        if 'conn' in locals() and conn is not None:
            conn.close()

        # Delete the file if it exists
        if os.path.exists(path_db):
            os.remove(path_db)

        # Raise the exception to propagate the error
        raise err
    
    finally:
        # Committing the changes
        conn.commit()
        # Closing the database connection
        conn.close()



import logging
from fluxo import fluxo_core

logging.basicConfig(
    format='%(asctime)s - %(levelname)s ===> %(message)s',
    datefmt='%Y/%m/%d %H:%M:%S',
    level=logging.INFO
)

logger = logging.getLogger(fluxo_core.__name__)




class Minutes:
    '''
    A class representing time in minutes and seconds.

    Args:
        - minutes (int): The number of minutes.
        - seconds (int): The number of seconds (0 to 59).

    Attributes:
        - minutes (int): The number of minutes.
        - seconds (str): The formatted representation of seconds as a two-digit string.

    Methods:
        - format(): Returns a dictionary with 'minutes' and 'at' keys representing the time.

    Example:
    >>> time = Minutes(5, 30)
    >>> time.format()
    {'minutes': 5, 'at': ':30'}
    '''
    def __init__(self, minutes: int, seconds: int) -> None:
        if not isinstance(minutes, int):
            raise ValueError("Minutes must be an integer.")
        
        if not (isinstance(seconds, int) and 0 <= seconds <= 59):
            raise ValueError("Seconds must be a int representing a two-digit integer between 0 and 59.")
        
        if not seconds >= 10:
            seconds = ''.join(['0',str(seconds)])

        self.minutes = minutes
        self.seconds = seconds

    def format(self):
        return {'minutes': self.minutes, 'at': f':{self.seconds}'}
    

class Hours:
    '''
    A class representing time in hours and minutes.

    Args:
        - hours (int): The number of hours.
        - minutes (int): The number of minutes (0 to 59).

    Attributes:
        - hours (int): The number of hours.
        - minutes (str): The formatted representation of minutes as a two-digit string.

    Methods:
        - format(): Returns a dictionary with 'hours' and 'at' keys representing the time.

    Example:
    >>> time = Hours(3, 45)
    >>> time.format()
    {'hours': 3, 'at': ':45'}
    '''
    
    def __init__(self, hours: int, minutes: int) -> None:
        if not isinstance(hours, int):
            raise ValueError("Hours must be an integer.")
        
        if not (isinstance(minutes, int) and 0 <= minutes <= 59):
            raise ValueError("Minutes must be a int representing a two-digit integer between 0 and 59.")
        
        if not minutes >= 10:
            minutes = ''.join(['0',str(minutes)])

        self.hours = hours
        self.minutes = minutes

    def format(self):
        return {'hours': self.hours, 'at': f':{self.minutes}'}
    

class Days:
    '''
    A class representing time in days, hours, and minutes.

    Args:
        - days (int): The number of days.
        - hours_minutes (tuple): A tuple representing hours and minutes (0 to 23 and 0 to 59, respectively).
    
    Attributes:
        - days (int): The number of days.
        - hours (str): The formatted representation of hours as a two-digit string.
        - minutes (str): The formatted representation of minutes as a two-digit string.

    Methods:
    - format(): Returns a dictionary with 'days' and 'at' keys representing the time.

    Example:
    >>> time = Days(5, (18, 30))
    >>> time.format()
    {'days': 5, 'at': '18:30'}
    '''
    def __init__(self, days: int, hours_minutes: tuple) -> None:
        hours = hours_minutes[0]
        minutes = hours_minutes[1]

        if not isinstance(days, int):
            raise ValueError("Days must be an integer.")
        
        if not (isinstance(hours_minutes, tuple) and 0 <= hours <= 23 and 0 <= minutes <= 59):
            raise ValueError("Hours and Minutes must be a tuple representing hours and minutes between (0,0) and (23,59).")
        
        if not (isinstance(hours, int) and isinstance(minutes, int)):
            raise ValueError('Values in tuple must be integer.')
        
        if not hours >= 10:
            hours = ''.join(['0',str(hours)])

        if not minutes >= 10:
            minutes = ''.join(['0',str(minutes)])

        self.days = days
        self.hours = hours
        self.minutes = minutes

    def format(self):
        return {'days': self.days, 'at': f'{self.hours}:{self.minutes}'}


import sqlite3
import json
from datetime import datetime
from dataclasses import dataclass
from fluxo.settings import Db
from fluxo.uttils import current_time_formatted


@dataclass
class ModelLogExecutionFlow:
    '''
    Represents the log of flow execution, storing information about tasks, errors, and execution times.

    Attributes:
        id (int, optional): The unique identifier for the log entry.
        name (str, optional): The name of the flow associated with the log entry.
        date_of_creation (datetime, optional): The date and time when the log entry was created.
        start_time (datetime, optional): The start time of the flow execution.
        end_time (datetime, optional): The end time of the fluxo execution.
        id_flow (int, optional): The identifier of the associated flow.
        ids_task (list, optional): A list of task IDs involved in the flow.
        ids_error_task (list, optional): A list of task IDs that encountered errors during execution.

    Methods:
        save(self): Save the log entry to the database and return the updated LogExecutionFluxo instance.
        update(cls, id, name, date_of_creation, start_time, end_time, id_fluxo, ids_task, ids_error_task):
            Update an existing log entry in the database and return the updated LogExecutionFluxo instance.
        get_all(cls): Retrieve all log entries from the database.
        get_by_name(cls, name): Retrieve a log entry by its associated fluxo name from the database.
        get_by_id(cls, id): Retrieve a log entry by its unique identifier from the database.
        get_by_idflow_and_endtime_is_none(cls, id_flow): Retrieve a log entry for a specific fluxo
            where the end time is not set.
        get_all_by_id_flow(cls, id_flow): Retrieve all log entries for a specific flow from the database.
        delete(cls, id): Delete a log entry by its unique identifier from the database.
    '''
    id: int = None
    name: str = None
    date_of_creation: datetime = None
    start_time: datetime = None
    end_time: datetime = None
    id_flow: int = None
    ids_task: list = None
    ids_error_task: list = None

    def save(self):
        date_of_creation = current_time_formatted()
        ids_task = json.dumps(self.ids_task) if self.ids_task else None
        ids_error_task = json.dumps(self.ids_error_task) if self.ids_error_task else None

        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('''
            INSERT INTO TB_LogExecutionFlow (name, date_of_creation, start_time, end_time, id_flow, ids_task, ids_error_task)
            VALUES (?, ?, ?, ?, ?, ?, ?)
        ''', (self.name, date_of_creation, self.start_time, self.end_time, self.id_flow, ids_task, ids_error_task))
        conn.commit()

        cursor.execute('SELECT last_insert_rowid()')
        id_log_execution_flow = cursor.fetchone()[0]

        conn.close()

        return ModelLogExecutionFlow.get_by_id(id_log_execution_flow)

    @staticmethod
    def update(id, name, date_of_creation, start_time, end_time, id_flow, ids_task, ids_error_task):
        ids_task = json.dumps(ids_task) if ids_task else None
        ids_error_task = json.dumps(ids_error_task) if ids_error_task else None

        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('''
            UPDATE TB_LogExecutionFlow
            SET name=?, date_of_creation=?, start_time=?, end_time=?, id_flow=?, ids_task=?, ids_error_task=?
            WHERE id=?
        ''', (name, date_of_creation, start_time, end_time, id_flow, ids_task, ids_error_task, id))
        conn.commit()

        conn.close()

        return ModelLogExecutionFlow.get_by_id(id)

    @staticmethod
    def get_all():
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_LogExecutionFlow')
        data = cursor.fetchall()
        conn.close()

        log_flows = []
        for row in data:
            log_flow = ModelLogExecutionFlow(*row)
            
            # Converter strings JSON from ids_task and ids_error_task to lists
            log_flow.ids_task = json.loads(log_flow.ids_task) if log_flow.ids_task else None
            log_flow.ids_error_task = json.loads(log_flow.ids_error_task) if log_flow.ids_error_task else None

            log_flows.append(log_flow)

        if log_flows:
            return log_flows
        else:
            return None

    @staticmethod
    def get_by_name(name):
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_LogExecutionFlow WHERE name=?', (name,))
        data = cursor.fetchone()
        conn.close()

        if data:
            log_flow = ModelLogExecutionFlow(*data)

            # Converter strings JSON from ids_task and ids_error_task to lists
            log_flow.ids_task = json.loads(log_flow.ids_task) if log_flow.ids_task else None
            log_flow.ids_error_task = json.loads(log_flow.ids_error_task) if log_flow.ids_error_task else None

            return log_flow
        else:
            return None

    @staticmethod
    def get_by_id(id):
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_LogExecutionFlow WHERE id=?', (id,))
        data = cursor.fetchone()
        conn.close()

        if data:
            log_flow = ModelLogExecutionFlow(*data)

            # Converter strings JSON from ids_task and ids_error_task to lists
            log_flow.ids_task = json.loads(log_flow.ids_task) if log_flow.ids_task else None
            log_flow.ids_error_task = json.loads(log_flow.ids_error_task) if log_flow.ids_error_task else None

            return log_flow
        else:
            return None
        
    @staticmethod
    def get_by_idflow_and_endtime_is_none(id_flow):
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_LogExecutionFlow WHERE id_flow=? AND end_time IS NULL', (id_flow,))
        data = cursor.fetchone()
        conn.close()

        if data:
            log_flow = ModelLogExecutionFlow(*data)

            # Converter strings JSON from ids_task and ids_error_task to lists
            log_flow.ids_task = json.loads(log_flow.ids_task) if log_flow.ids_task else None
            log_flow.ids_error_task = json.loads(log_flow.ids_error_task) if log_flow.ids_error_task else None

            return log_flow
        else:
            return None

    @staticmethod
    def get_all_by_id_flow(id_flow):
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_LogExecutionFlow WHERE id_flow=?', (id_flow,))
        data = cursor.fetchall()
        conn.close()

        log_flows = []
        for row in data:
            log_flow = ModelLogExecutionFlow(*row)
            
            # Converter strings JSON from ids_task and ids_error_task to lists
            log_flow.ids_task = json.loads(log_flow.ids_task) if log_flow.ids_task else None
            log_flow.ids_error_task = json.loads(log_flow.ids_error_task) if log_flow.ids_error_task else None

            log_flows.append(log_flow)

        if log_flows:
            return log_flows
        else:
            return None

    @staticmethod
    def delete(id):
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('DELETE FROM TB_LogExecutionFlow WHERE id=?', (id,))
        conn.commit()
        conn.close()

    def __repr__(self) -> str:
        '''
        Returns a string representation of the 'LogExecutionFlow' instance.
        '''
        return f'''
            id:                     {self.id},
            name:                   {self.name},
            date_of_creation:       {self.date_of_creation},
            start_time:             {self.start_time},
            end_time:               {self.end_time},
            id_flow:                {self.id_flow},
            ids_task:               {self.ids_task},
            ids_error_task:         {self.ids_error_task},
        '''



import flet as ft
from fluxo.settings import AppThemeColors


class AppBar(ft.UserControl):
    def __init__(sefl):
        super().__init__()

    def build(self):
        return ft.Container(
            content=ft.Row(
                controls=[
                    ft.Container(
                        content=ft.Row(
                            controls=[
                                ft.Image(
                                    src='icons/icon-512.png',
                                    repeat=ft.ImageRepeat.NO_REPEAT,
                                    fit=ft.ImageFit.CONTAIN,
                                    width=40,
                                    height=40,
                                ),
                                ft.Text(
                                    value='Fluxo',
                                    weight=ft.FontWeight.BOLD,
                                    color=AppThemeColors.BLACK,
                                    size=25
                                ), # Text
                            ]
                        ),
                        on_click=self.on_click_container_home,
                    )
                ],
            ),
            padding=ft.padding.only(left=10, top=5, right=10, bottom=0),
        )
    
    async def on_click_container_home(self, e):
        await self.page.go_async('/')


import sqlite3
from datetime import datetime
from dataclasses import dataclass
from fluxo.settings import Db
from fluxo.uttils import current_time_formatted


@dataclass
class ModelTask:
    '''
    Represents a 'Task' object with attributes corresponding to the columns
    in the 'TB_Task' table in the SQLite database.

    Attributes:
    - id (int): The unique identifier for the 'Task'.
    - name (str): The name of the 'Task'.
    - execution_date (datetime): The date and time when the 'Task' was executed.
    - flow_id (int): The ID of the 'Flow' associated with the 'Task'.
    - start_time (datetime): The start time of the 'Task'.
    - end_time (datetime): The end time of the 'Task'.
    - error (str): Any error message associated with the 'Task'.

    Methods:
    - save(): Saves the current 'Task' instance to the 'TB_Task' table in the database.
    - update(id, name, execution_date, flow_id, start_time, end_time, error): Updates the 'Task' with the specified ID
      with the provided information in the 'TB_Task' table.
    - get_all(): Retrieves all 'Task' instances from the 'TB_Task' table.
    - get_by_name(name): Retrieves a 'Task' instance by its name from the 'TB_Task' table.
    - get_by_id(id): Retrieves a 'Task' instance by its ID from the 'TB_Task' table.
    - get_all_by_flow_id(flow_id): Retrieves all 'Task' instances associated with the specified 'Flow' ID.
    - delete(id): Deletes the 'Task' with the specified ID from the 'TB_Task' table.
    - __repr__(): Returns a string representation of the 'Task' instance.
    '''
    id: int = None
    name: str = None
    execution_date: datetime = None
    flow_id: int = None
    start_time: datetime = None
    end_time: datetime = None
    error: str = None

    def save(self):
        '''
        Saves the current 'Task' instance to the 'TB_Task' table in the database.

        Returns:
            Task: The saved 'Task' instance.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('''
            INSERT INTO TB_Task (name, execution_date, flow_id, start_time, end_time, error)
            VALUES (?, ?, ?, ?, ?, ?)
        ''', (self.name, self.execution_date, self.flow_id, self.start_time, self.end_time, self.error))
        conn.commit()

        # Retrieve task ID after insertion
        cursor.execute('SELECT last_insert_rowid()')
        task_id = cursor.fetchone()[0]

        conn.close()

        return ModelTask.get_by_id(task_id)

    @staticmethod
    def update(id, name, execution_date, flow_id, start_time, end_time, error):
        '''
        Updates the 'Task' with the specified ID with the provided information
        in the 'TB_Task' table.

        Parameters:
        - id (int): The ID of the 'Task' to be updated.
        - name (str): The new name for the 'Task'.
        - execution_date (datetime): The new execution date and time for the 'Task'.
        - fllow_id (int): The new 'Fluxo' ID associated with the 'Task'.
        - start_time (datetime): The new start time for the 'Task'.
        - end_time (datetime): The new end time for the 'Task'.
        - error (str): The new error message for the 'Task'.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('''
            UPDATE TB_Task
            SET name=?, execution_date=?, flow_id=?, start_time=?, end_time=?, error=?
            WHERE id=?
        ''', (name, execution_date, flow_id, start_time, end_time, error, id))
        conn.commit()
        conn.close()

    @staticmethod
    def get_all():
        '''
        Retrieves all 'Task' instances from the 'TB_Task' table.

        Returns:
            List[Task] or None: A list containing all 'Task' instances in the database, or None if no tasks are found.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_Task')
        data = cursor.fetchall()
        conn.close()
        if data:
            return [ModelTask(*row) for row in data]
        else:
            return None

    @staticmethod
    def get_by_name(name):
        '''
        Retrieves a 'Task' instance by its name from the 'TB_Task' table.

        Parameters:
            - name (str): The name of the 'Task' to be retrieved.

        Returns:
            Task or None: The 'Task' instance if found, or None if not found.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_Task WHERE name=?', (name,))
        data = cursor.fetchone()
        conn.close()
        if data:
            return ModelTask(*data)
        else:
            return None

    @staticmethod
    def get_by_id(id):
        '''
        Retrieves a 'Task' instance by its ID from the 'TB_Task' table.

        Parameters:
            - id (int): The ID of the 'Task' to be retrieved.

        Returns:
            Task or None: The 'Task' instance if found, or None if not found.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_Task WHERE id=?', (id,))
        data = cursor.fetchone()
        conn.close()
        if data:
            return ModelTask(*data)
        else:
            return None

    @staticmethod
    def get_all_by_fluxo_id(flow_id):
        '''
        Retrieves all 'Task' instances associated with the specified 'Fllow' ID from the 'TB_Task' table.

        Parameters:
            - flow_id (int): The ID of the 'Flow' for which tasks are to be retrieved.

        Returns:
            List[Task] or None: A list containing all 'Task' instances associated with the specified 'Flow' ID,
                or None if no tasks are found.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_Task WHERE flow_id=?', (flow_id,))
        data = cursor.fetchall()
        conn.close()
        if data:
            return [ModelTask(*row) for row in data]
        else:
            return None

    @staticmethod
    def delete(id):
        '''
        Deletes the 'Task' with the specified ID from the 'TB_Task' table.

        Parameters:
            - id (int): The ID of the 'Task' to be deleted.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('DELETE FROM TB_Task WHERE id=?', (id,))
        conn.commit()
        conn.close()

    def __repr__(self) -> str:
        '''
        Returns a string representation of the 'Task' instance.
        '''
        return f'''
            id:                     {self.id},
            name:                   {self.name},
            execution_date:         {self.execution_date},
            flow_id:                {self.flow_id},
            start_time:             {self.start_time},
            end_time:               {self.end_time},
            error:                  {self.error},
        '''



import flet as ft
import asyncio
from fluxo.settings import AppThemeColors
from fluxo.fluxo_server.screens.home.flow import Flow
from fluxo.fluxo_server.screens.app_bar import AppBar
from fluxo.fluxo_server.screens.footer import Footer
from fluxo.fluxo_core.flows_executor import FlowsExecutor
from fluxo.fluxo_core.database.flow import ModelFlow
from fluxo.fluxo_core.database.app import ModelApp


class Home(ft.UserControl):
    def __init__(self):
        super().__init__()
        self.flows_executor = FlowsExecutor()

    def build(self):
        self.column_flows = ft.Ref[ft.Column]()
        self.responsiverow_flows = ft.Ref[ft.ResponsiveRow]()

        return ft.Column(
            ref=self.column_flows,
            controls=[
                ft.Container(height=0),
                ft.Container(
                    content=ft.Row(
                        controls=[
                            ft.Text(
                                value='Flows',
                                weight=ft.FontWeight.BOLD,
                                color=AppThemeColors.BLACK,
                                size=25,
                                expand=True
                            ), # Text
                            ft.Tooltip(
                                message='Update new Flow',
                                content=ft.FloatingActionButton(
                                    content=ft.Text(
                                        value='Update new Flow', 
                                        weight=ft.FontWeight.BOLD,
                                        color=AppThemeColors.WHITE
                                    ),
                                    bgcolor=AppThemeColors.PRIMARY,
                                    width=140,
                                    height=27,
                                    on_click=self.on_click_floatingactionbutton_update_new_flow
                                ),
                                bgcolor=AppThemeColors.QUARTENARY
                            ),
                            ft.Tooltip(
                                message='Synchronize',
                                content=ft.FloatingActionButton(
                                    content=ft.Icon(
                                        name=ft.icons.SYNC,
                                        color=AppThemeColors.WHITE
                                    ),
                                    bgcolor=AppThemeColors.PRIMARY,
                                    width=50,
                                    height=27,
                                    on_click=self.on_click_floatingactionbutton_sync
                                ),
                                bgcolor=AppThemeColors.QUARTENARY
                            ),
                        ]
                    ),
                    width=900
                ), # Container
                ft.Container(
                    content=ft.ResponsiveRow(
                        ref=self.responsiverow_flows,
                        controls=[

                        ], # controls
                        alignment=ft.MainAxisAlignment.CENTER
                    ), # ResponsiveRow
                    alignment=ft.alignment.center,
                    width=900
                ), # Container
            ], # controls
            scroll=ft.ScrollMode.AUTO,
            horizontal_alignment=ft.CrossAxisAlignment.CENTER
        ) # Column
    
    async def _load_flows(self):
        self.responsiverow_flows.current.controls.clear()
        flows = ModelFlow.get_all()
        if flows:
            for flow in flows:
                flow = ModelFlow.get_by_name(flow.name)
                self.responsiverow_flows.current.controls.append(Flow(flow=flow, flows_executor=self.flows_executor))
                await self.update_async()

    async def on_click_floatingactionbutton_update_new_flow(self, e):
        self.flows_executor.update_new_flow_in_python_files()

        await self.clean_async()
        await self._load_flows()

    async def on_click_floatingactionbutton_sync(self, e):
        await self.clean_async()
        await self.did_mount_async()

    async def did_mount_async(self):
        self.task_load_flows = asyncio.create_task(self._load_flows())

    async def will_unmount_async(self):
        self.task_load_flows.cancel()


def view_home():
    home = Home()

    return ft.View(
        route='/',
        controls=[
            AppBar(),
            ft.Container(
                content=home,
                expand=True,
                padding=ft.padding.all(15)
            ),
            Footer()
        ],
        padding=ft.padding.all(0),
        bgcolor=AppThemeColors.WHITE,
    )


import sqlite3
from datetime import datetime
from dataclasses import dataclass
from fluxo.settings import Db
from fluxo.uttils import current_time_formatted


@dataclass
class ModelApp:
    '''
    Represents an application status in database

    Attributes:
        id (int): The unique identifier for the application.
        active (bool): Indicates whether the application is currently active.
        active_since (datetime): The datetime when the application was last activated.

    Methods:
        save(self): Saves the current state of the application to the database.
        update(cls, id: int, active: bool): Updates the activation status of the application in the database.
        get(cls, id: int = 1): Retrieves an application instance based on the provided ID from the database.
        __repr__(self): Returns a string representation of the 'App' instance.
    '''
    id: int = None
    active: bool = None
    active_since: datetime = None

    def save(self):
        '''
        Saves the current state of the application to the database.

        Returns:
            App: An instance of the 'App' class.
        '''
        if self.active:
            self.active_since = current_time_formatted()
        else:
            self.active_since = None
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('''
            INSERT INTO TB_App (active, active_since)
            VALUES (?, ?)
        ''', (self.active, self.active_since))
        conn.commit()

        cursor.execute('SELECT last_insert_rowid()')
        task_id = cursor.fetchone()[0]

        conn.close()

        return ModelApp.get()

    @staticmethod
    def update(id: int, active: bool):
        '''
        Updates the activation status of the application in the database.

        Args:
            id (int): The unique identifier of the application to be updated.
            active (bool): The new activation status for the application.
        '''
        active_since: datetime
        if active:
            active_since = current_time_formatted()
        else:
            active_since = None
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('''
            UPDATE TB_App
            SET active=?, active_since=?
            WHERE id=?
        ''', (active, active_since, id))
        conn.commit()
        conn.close()

        return ModelApp.get()

    @staticmethod
    def get(id: int = 1):
        '''
        Retrieves an application instance based on the provided ID from the database.

        Args:
            id (int): The unique identifier of the application to be retrieved.

        Returns:
            App or None: An instance of the 'App' class if found, else None.
        '''
        conn = sqlite3.connect(Db.PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT * FROM TB_App WHERE id=?', (id,))
        data = cursor.fetchone()
        conn.close()
        if data:
            return ModelApp(*data)
        else:
            return None


    def __repr__(self) -> str:
        '''
        Returns a string representation of the 'ModelApp' instance.
        '''
        return f'''
            id:                   {self.id},
            active:               {self.active},
            active_since:         {self.active_since},
        '''



import flet as ft
import asyncio
from fluxo.settings import AppThemeColors
from fluxo.fluxo_server.screens.app_bar import AppBar
from fluxo.fluxo_server.screens.footer import Footer
from fluxo.fluxo_core.database.flow import ModelFlow
from fluxo.fluxo_core.database.task import ModelTask
from fluxo.fluxo_core.database.log_execution_flow import ModelLogExecutionFlow


class Task(ft.UserControl):
    def __init__(self, log_flow: ModelLogExecutionFlow, name_task: str):
        super().__init__()
        self.log_flow = log_flow
        self.name_task = name_task

    def build(self):
        self.container_task = ft.Ref[ft.Container]()
        self.container_status_execution = ft.Ref[ft.Container]()
        self.text_name_task = ft.Ref[ft.Text]()
        self.text_start_end_execution = ft.Ref[ft.Text]()

        return ft.Container(
            ref=self.container_task,
            content=ft.Row(
                controls=[
                    ft.Container(
                        ref=self.container_status_execution,
                        height=15,
                        width=15,
                        border_radius=ft.border_radius.all(15),
                    ), # Container
                    ft.Text(
                        ref=self.text_name_task,
                        color=AppThemeColors.BLACK,
                        size=15,
                        expand=True
                    ),
                    ft.Text(
                        ref=self.text_start_end_execution,
                        color=AppThemeColors.BLACK,
                        size=10,
                    )
                ]
            ),
            padding=ft.padding.all(15),
            bgcolor=AppThemeColors.GREY,
            border_radius=ft.border_radius.all(10),
            height=50
        )
    
    async def _load_task(self):
        tasks_flow = ModelTask.get_all_by_fluxo_id(self.log_flow.id_flow)
        for task in tasks_flow:
            if self.log_flow.ids_task:
                if task.name == self.name_task and task.id in self.log_flow.ids_task:
                    await self._load_container_status_execution(task)
                    self.container_task.current.on_click = self._on_click_task
                    self.text_name_task.current.value = task.name
                    self.text_start_end_execution.current.value = f'({task.start_time} - {task.end_time})'
                    self.task_id = task.id
                else:
                    if task.name == self.name_task and task.end_time is None:
                        await self._load_container_status_execution(task)
                        self.container_task.current.on_click = self._on_click_task
                        self.text_name_task.current.value = task.name
                        self.text_start_end_execution.current.value = f'({task.start_time} - {task.end_time})'
                        self.task_id = task.id
            else:
                if task.end_time is None and task.name == self.name_task:
                    await self._load_container_status_execution(task)
                    self.container_task.current.on_click = self._on_click_task
                    self.text_name_task.current.value = task.name
                    self.text_start_end_execution.current.value = f'({task.start_time} - {task.end_time})'
                    self.task_id = task.id
                else:
                    self.container_task.current.on_click = None
                    self.container_status_execution.current.bgcolor = AppThemeColors.WHITE
                    self.text_name_task.current.value = self.name_task
                    self.text_start_end_execution.current.value = 'Execution not yet started'

            await self.update_async()

    async def _load_container_status_execution(self, task):
        if task.end_time is None:
            self.container_status_execution.current.bgcolor = AppThemeColors.BLUE
        elif task.end_time and task.error:
            self.container_status_execution.current.bgcolor = AppThemeColors.RED
        else:
            self.container_status_execution.current.bgcolor = AppThemeColors.GREEN

    async def _on_click_task(self, e):
        await self.page.go_async(f'task/{self.task_id}')

    async def did_mount_async(self):
        self.task_load_log_flow = asyncio.create_task(self._load_task())

    async def will_unmount_async(self):
        self.task_load_log_flow.cancel()


import flet as ft
import asyncio
from fluxo.settings import AppThemeColors
from fluxo.fluxo_core.database.log_execution_flow import ModelLogExecutionFlow


class StatusExecution(ft.UserControl):
    def __init__(self, log_flow: ModelLogExecutionFlow):
        super().__init__()
        self.log_flow = log_flow

    def build(self):
        self.container_execution = ft.Ref[ft.Container]()
        self.tooltip_execution = ft.Ref[ft.Tooltip]()
    
        return ft.Tooltip(
            ref=self.tooltip_execution,
            message='Synchronize',
            content=ft.Container(
                ref=self.container_execution,
                height=23,
                width=23,
                border_radius=ft.border_radius.all(15),
                on_click=self.on_click_log_flow,
            ),
            bgcolor=AppThemeColors.QUARTENARY
        )
    
    async def on_click_log_flow(self, e):
        await self.page.go_async(f'flow-execution/{self.log_flow.id}')
    
    async def _load_status_execution(self):        
        if self.log_flow.end_time is None:
            self.container_execution.current.bgcolor = AppThemeColors.BLUE
        elif self.log_flow.end_time and self.log_flow.ids_error_task:
            self.container_execution.current.bgcolor = AppThemeColors.RED
            self.tooltip_execution.current.message = self.log_flow.end_time
        else:
            self.container_execution.current.bgcolor = AppThemeColors.GREEN
            self.tooltip_execution.current.message = self.log_flow.end_time

        await self.update_async()

    async def did_mount_async(self):
        self.flow_load_status_execution = asyncio.create_task(self._load_status_execution())

    async def will_unmount_async(self):
        self.flow_load_status_execution.cancel()


from fluxo.settings import AppSettings

from fluxo.fluxo_core.flow import Flow
from fluxo.fluxo_core.task import Task

from fluxo.fluxo_core.intervals import (
    Minutes,
    Hours,
    Days
)

__version__ = AppSettings.VERSION


import flet as ft
import asyncio
from fluxo.settings import AppThemeColors
from fluxo.fluxo_server.screens.app_bar import AppBar
from fluxo.fluxo_server.screens.footer import Footer
from fluxo.fluxo_server.screens.flow_execution.task import Task
from fluxo.fluxo_core.database.flow import ModelFlow
from fluxo.fluxo_core.database.task import ModelTask
from fluxo.fluxo_core.database.log_execution_flow import ModelLogExecutionFlow


class FlowExecution(ft.UserControl):
    def __init__(self, log_flow_id):
        super().__init__()
        self.log_flow_id = log_flow_id

    def build(self):
        self.text_name_flow = ft.Ref[ft.Text]()
        self.text_start_end_execution = ft.Ref[ft.Text]()
        self.container_status_execution = ft.Ref[ft.Container]()
        self.responsiverow_tasks = ft.Ref[ft.ResponsiveRow]()

        return ft.Column(
            controls=[
                ft.Container(
                    content=ft.ResponsiveRow(
                        controls=[
                            ft.Row(
                                controls=[
                                    ft.IconButton(
                                        icon=ft.icons.ARROW_BACK_ROUNDED,
                                        on_click=self.iconbutton_go_back,
                                        icon_color=AppThemeColors.BLACK,
                                        icon_size=30
                                    ),
                                    ft.Text(
                                        ref=self.text_name_flow,
                                        weight=ft.FontWeight.BOLD,
                                        color=AppThemeColors.BLACK,
                                        size=20
                                    ), # Text
                                    ft.Text(' > ', size=20, color=AppThemeColors.BLACK_SECONDARY), # Text
                                    ft.Text(
                                        ref=self.text_start_end_execution,
                                        color=AppThemeColors.BLACK,
                                        size=15
                                    ), # Text
                                    ft.Container(
                                        ref=self.container_status_execution,
                                        height=15,
                                        width=15,
                                        border_radius=ft.border_radius.all(15),
                                    ), # Container
                                ], # controls
                            ), # Row
                            ft.Container(
                                content=ft.ResponsiveRow(
                                    ref=self.responsiverow_tasks,
                                    controls=[
                                        
                                    ]
                                ),
                                padding=ft.padding.only(left=50, top=15, right=20, bottom=15)
                            ), # Container
                        ], # controls
                    ), # ResponsiveRow
                    width=900
                ), # Container
                ft.Container(height=0, padding=ft.padding.all(0)),
            ], # controls
            scroll=ft.ScrollMode.AUTO,
            horizontal_alignment=ft.CrossAxisAlignment.CENTER,
            spacing=0
        )
    
    async def iconbutton_go_back(self, e):
        await self.page.go_async('/')

    async def _load_log_flow(self):
        log_flow = ModelLogExecutionFlow.get_by_id(self.log_flow_id)

        # Update name and date execution
        self.text_name_flow.current.value = log_flow.name
        self.text_start_end_execution.current.value = f'({log_flow.date_of_creation} - {log_flow.end_time})'
        
        # Update color status
        if log_flow.end_time is None:
            self.container_status_execution.current.bgcolor = AppThemeColors.BLUE
        elif log_flow.end_time and log_flow.ids_error_task:
            self.container_status_execution.current.bgcolor = AppThemeColors.RED
        else:
            self.container_status_execution.current.bgcolor = AppThemeColors.GREEN

        # Update tasks
        flow = ModelFlow.get_by_id(log_flow.id_flow)
        name_tasks_flow = flow.list_names_tasks

        for name_task in name_tasks_flow:
            self.responsiverow_tasks.current.controls.append(
                Task(log_flow, name_task)
            )
            await self.update_async()

        self.page.session.set('log_flow_id', self.log_flow_id)

    async def did_mount_async(self):
        self.task_load_log_flow = asyncio.create_task(self._load_log_flow())

    async def will_unmount_async(self):
        self.task_load_log_flow.cancel()


def view_flow_execution(log_flow_id: int):

    return ft.View(
        route='flow-execution/:id',
        controls=[
            AppBar(),
            ft.Container(
                content=FlowExecution(log_flow_id),
                expand=True,
                padding=ft.padding.only(left=15, top=0, right=15, bottom=15)
            ),
            Footer()
        ],
        padding=ft.padding.all(0),
        bgcolor=AppThemeColors.WHITE,
    )


from fluxo.fluxo_core.flows_executor import FlowsExecutor
from fluxo.logging import logger


if __name__ == '__main__':
    flows_executor = FlowsExecutor()
    try:
        flows_executor.execute_parallel_flows()
    except KeyboardInterrupt:
        logger.warning('Program interrupted by the user')


import traceback
from fluxo.logging import logger
from datetime import datetime
from fluxo.fluxo_core.flow import Flow
from fluxo.fluxo_core.database.flow import ModelFlow
from fluxo.fluxo_core.database.task import ModelTask
from fluxo.fluxo_core.database.log_execution_flow import ModelLogExecutionFlow
from fluxo.uttils import current_time_formatted, convert_str_to_datetime, convert_datetime_to_str


class Task:
    '''
    Represents a task in a workflow, enabling the execution of asynchronous functions
    with logging and error handling.

    Parameters:
        - name (str): The name of the task.
        - flow (Flow): The associated flow to which the task belongs.
        - start_time (datetime, optional): The start time of the task.
        - end_time (datetime, optional): The end time of the task.

    Example:
        ```
        from fluxo import Flow, Task, Minutes

        interval = Minutes(1, 30).format()
        flow = Flow(name='My Flow 1', interval=interval)

        @Task('My Task 1', flow=flow)
        async def My_func():
            print('My_func executed!')
        ```

    '''
    def __init__(
        self,
        name: str,
        flow: Flow,
        start_time: datetime = None,
        end_time: datetime = None
    ):
        self.task_info = {
            'name': name,
            'flow': flow,
            'start_time': start_time,
            'end_time': end_time,
        }

    def __call__(self, func):
        '''
        Decorates an asynchronous function to enable logging and error handling.

        Parameters:
            - func: The asynchronous function to be decorated.

        Returns:
            - wrapper: The decorated asynchronous function.

        The decorated function is executed with logging, error handling, and updates to the database
        for tracking task execution.
        '''
        async def wrapper(*args, **kwargs):
            # Retrieve the 'Flow' information from the database
            flow_register_db = ModelFlow.get_by_name(
                self.task_info.get('flow').name)

            # Create a new 'ModelTask' instance and save it to the database
            task = ModelTask(name=self.task_info.get(
                'name'), flow_id=flow_register_db.id)
            new_task = task.save()

            # Params to update LogExecutionFlow
            _params = {
                'id_flow': flow_register_db.id,
                'id_task': new_task.id
            }

            try:
                # Call the original function
                new_task.start_time = current_time_formatted()
                new_task.update(**new_task.__dict__)
                self._newlog_execution_flow(**_params)

                # Function executed
                result = await func(*args, **kwargs)

                new_task.end_time = current_time_formatted()
                new_task.execution_date = new_task.end_time

                new_task.update(**new_task.__dict__)
                self._update_log_execution_flow(**_params)

                logger.info(f'Task [{new_task.name}] executed successfully')

                return result
            except Exception as err:
                error = traceback.format_exc()
                new_task.error = f'[Error in task: {new_task.name}]' + \
                    '\n' + error
                new_task.end_time = current_time_formatted()
                new_task.execution_date = new_task.end_time

                new_task.update(**new_task.__dict__)
                self._update_log_execution_flow(**_params)

                logger.info(f'Task [{new_task.name}] executed with error')

        setattr(wrapper, 'task_info', self.task_info)
        return wrapper
    
    def _newlog_execution_flow(self, **kwargs):
        '''
        Create the log of flow execution with task information.

        Parameters:
            - kwargs: Additional keyword arguments.

        This method create the log of flow execution in the database.
        '''
        log_flow = ModelLogExecutionFlow.get_by_idflow_and_endtime_is_none(kwargs['id_flow'])
        flow = ModelFlow.get_by_id(kwargs['id_flow'])

        # If LogExecutionFlow is not in the database, create a new instance and save it
        if log_flow is None:
            log_flow = ModelLogExecutionFlow(
                name=flow.name,
                id_flow=flow.id,
                start_time=current_time_formatted()
            )
            log_flow = log_flow.save()
    
    def _update_log_execution_flow(self, **kwargs):
        '''
        Updates the log of flow execution with task information.

        Parameters:
            - kwargs: Additional keyword arguments.

        This method updates the log of flow execution in the database
        with information about the associated task.
        '''
        log_flow = ModelLogExecutionFlow.get_by_idflow_and_endtime_is_none(kwargs['id_flow'])
        flow = ModelFlow.get_by_id(kwargs['id_flow'])
        task = ModelTask.get_by_id(kwargs['id_task'])

        # When tasks by log_flow is None, update the log with the current task information
        if log_flow.ids_task is None:
            log_flow.ids_task = [task.id]
            if log_flow.start_time is None:
                log_flow.start_time = task.start_time
            log_flow = log_flow.update(**log_flow.__dict__)
        else:
            # When the number of tasks in the log is less than the total tasks in the fluxo,
            # update the log with the current task information
            if len(log_flow.ids_task) < len(flow.list_names_tasks):
                log_flow.ids_task.append(task.id)
                log_flow = log_flow.update(**log_flow.__dict__)

        # When all tasks in the flow are completed, update the log with end time and handle errors
        if len(log_flow.ids_task) == len(flow.list_names_tasks):
            list_end_time_tasks = []
            for id_task in log_flow.ids_task:
                task = ModelTask.get_by_id(id_task)
                list_end_time_tasks.append(convert_str_to_datetime(task.end_time))

                # If a task has an error, update the error log in the flow log
                if task.error:
                    if log_flow.ids_error_task is None:
                        log_flow.ids_error_task = [task.id]
                    else:
                        log_flow.ids_error_task.append(task.id)

            log_flow.end_time = convert_datetime_to_str(max(list_end_time_tasks))
            log_flow.update(**log_flow.__dict__)
